#!/usr/bin/python3

# Copyright (c) 2017 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

DOCUMENTATION = '''
---
module: kolla_passwords
short_description: Generates a kolla-ansible passwords file
'''

from ansible.module_utils.basic import *

import os.path
import shutil
import tempfile

IMPORT_ERRORS = []
try:
    import yaml
except ImportError as e:
    IMPORT_ERRORS.append(e)


def virtualenv_path_prefix(module):
    return "%s/bin" % module.params['virtualenv']


def kolla_genpwd(module, file_path):
    """Run the kolla-genpwd command."""
    cmd = ["kolla-genpwd", "--passwords", file_path]
    module.run_command(cmd, check_rc=True,
                       path_prefix=virtualenv_path_prefix(module))


def kolla_mergepwd(module, old_path, new_path, final_path):
    """Run the kolla-mergepwd command."""
    cmd = ["kolla-mergepwd",
           "--old", old_path,
           "--new", new_path,
           "--final", final_path]
    module.run_command(cmd, check_rc=True,
                       path_prefix=virtualenv_path_prefix(module))


def create_vault_password_file(module):
    """Create a vault password file."""
    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.write(module.params['vault_password'].encode())
        return f.name


def vault_encrypt(module, file_path):
    """Encrypt a file using Ansible vault"""
    password_path = create_vault_password_file(module)
    try:
        cmd = ["ansible-vault", "encrypt",
               "--vault-password-file", password_path,
               file_path]
        module.run_command(cmd, check_rc=True,
                           path_prefix=virtualenv_path_prefix(module))
    finally:
        os.unlink(password_path)


def vault_decrypt(module, file_path):
    """Decrypt a file using Ansible vault"""
    password_path = create_vault_password_file(module)
    try:
        cmd = ["ansible-vault", "decrypt",
               "--vault-password-file", password_path,
               file_path]
        module.run_command(cmd, check_rc=True,
                           path_prefix=virtualenv_path_prefix(module))
    finally:
        os.unlink(password_path)


def create_named_tempfile():
    """Create a named temporary file and return its name."""
    with tempfile.NamedTemporaryFile(delete=False) as f:
        temp_file_path = f.name
        f.close()
    return temp_file_path


def kolla_passwords(module):
    """Generate a kolla-ansible passwords.yml file.

    We use the sample passwords.yml file as a base to determine which passwords
    to generate. This gets merged with an existing passwords file if one
    exists.  We then apply any custom password overrides. Finally, we generate
    any passwords that are missing.  If requested, the final file will be
    encrypted using ansible vault.
    """
    if not os.path.isfile(module.params['sample']):
        module.fail_json(msg="Sample passwords.yml file %s does not exist" %
                         module.params['sample'])

    temp_file_path = create_named_tempfile()
    try:
        # Start with kolla's sample password file.
        shutil.copyfile(module.params['sample'], temp_file_path)

        # If passwords exist, decrypt and merge these in.
        if module.params['src'] and os.path.isfile(module.params['src']):
            src_path = create_named_tempfile()
            try:
                shutil.copyfile(module.params['src'], src_path)
                if module.params['vault_password']:
                    vault_decrypt(module, src_path)
                kolla_mergepwd(module, src_path, temp_file_path, temp_file_path)
            finally:
                os.unlink(src_path)

        # Merge in overrides.
        if module.params['overrides']:
            with tempfile.NamedTemporaryFile(delete=False) as f:
                # NOTE(mgoddard): Temporary files are opened in binary mode, so
                # specify an encoding.
                yaml.dump(module.params['overrides'], f, encoding='utf-8')
                overrides_path = f.name
            try:
                kolla_mergepwd(module, overrides_path, temp_file_path, temp_file_path)
            finally:
                os.unlink(overrides_path)

        # Generate null passwords.
        kolla_genpwd(module, temp_file_path)

        # Compare with the decrypted destination file.
        if os.path.isfile(module.params['dest']):
            if module.params['vault_password']:
                dest_path = create_named_tempfile()
                try:
                    shutil.copyfile(module.params['dest'], dest_path)
                    vault_decrypt(module, dest_path)
                    checksum_dest = module.sha1(dest_path)
                finally:
                    os.unlink(dest_path)
            else:
                checksum_dest = module.sha1(module.params['dest'])
            checksum_temp_file = module.sha1(temp_file_path)
            changed = checksum_dest != checksum_temp_file
        else:
            changed = True

        # Encrypt the file.
        if changed and module.params['vault_password']:
            vault_encrypt(module, temp_file_path)

        # Move into place.
        if changed and not module.check_mode:
            module.atomic_move(temp_file_path, module.params['dest'])
    except Exception as e:
        module.fail_json(msg="Failed to generate kolla passwords: %s" % repr(e))
    finally:
        if os.path.isfile(temp_file_path):
            os.unlink(temp_file_path)

    if not module.check_mode:
        # Update the file's attributes.
        file_args = module.load_file_common_arguments(module.params)
        changed = module.set_fs_attributes_if_different(file_args, changed)

    return {'changed': changed}


def main():
    module = AnsibleModule(
        argument_spec = dict(
            dest=dict(default='/etc/kolla/passwords.yml', type='str'),
            overrides=dict(default={}, type='dict'),
            sample=dict(default='/usr/share/kolla-ansible/etc_examples/kolla/passwords.yml', type='str'),
            src=dict(default='/etc/kolla/passwords.yml', type='str'),
            vault_password=dict(type='str', no_log=True),
            virtualenv=dict(type='str'),
        ),
        add_file_common_args=True,
        supports_check_mode=True,
    )

    if IMPORT_ERRORS:
        errors = ", ".join([repr(e) for e in IMPORT_ERRORS])
        module.fail_json(msg="Failed to import modules: %s" % errors)

    result = kolla_passwords(module)
    module.exit_json(**result)

if __name__ == '__main__':
    main()



# Copyright (c) 2020 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import unittest

import jinja2

from kayobe.plugins.action import kolla_ansible_host_vars


@jinja2.contextfilter
def _net_interface(context, name):
    return context.get(name + '_interface')


@jinja2.contextfilter
def _net_vlan(context, name):
    return context.get(name + '_vlan')


@jinja2.contextfilter
def _net_select_bridges(context, names):
    return [name for name in names
            if (_net_interface(context, name) or "").startswith("br")]


class FakeTemplar(object):

    def __init__(self, variables):
        self.variables = variables
        self.env = jinja2.Environment()
        self.env.filters['net_interface'] = _net_interface
        self.env.filters['net_vlan'] = _net_vlan
        self.env.filters['net_select_bridges'] = _net_select_bridges

    def template(self, string):
        template = self.env.from_string(string)
        result = template.render(**self.variables)
        return {
            "None": None,
            "True": True,
            "False": False,
        }.get(result, result)


class TestCase(unittest.TestCase):

    variables = {
        "network_interfaces": [
            "foo",
            "bar",
        ],
        "foo_interface": "eth0",
        "foo_vlan": 1,
        "bar_interface": "eth1",
        "bar_vlan": 2,
        "network_bridge_suffix_ovs": "-ovs",
        "network_patch_prefix": "p-",
        "network_patch_suffix_ovs": "-ovs",
    }

    def _create_module(self, variables=None):
        if not variables:
            variables = self.variables
        templar = FakeTemplar(variables)
        return kolla_ansible_host_vars.ActionModule(None, None, None, None,
                                                    templar, None)

    def test__run_empty_args(self):
        module = self._create_module()
        result = module._run([], [])
        expected = {
            "changed": False,
            "ansible_facts": {},
            "_ansible_facts_cacheable": False,
        }
        self.assertEqual(expected, result)

    def test__run_one_interface(self):
        module = self._create_module()
        interfaces = [{
            "var_name": "kolla_foo_interface",
            "network": "foo",
            "description": "Foo network",
            "required": False,
        }]
        result = module._run(interfaces, [])
        expected = {
            "changed": False,
            "ansible_facts": {
                "kolla_foo_interface": "eth0",
            },
            "_ansible_facts_cacheable": False,
        }
        self.assertEqual(expected, result)

    def test__run_two_interfaces(self):
        module = self._create_module()
        interfaces = [{
            "var_name": "kolla_foo_interface",
            "network": "foo",
            "description": "Foo network",
            "required": False,
        }, {
            "var_name": "kolla_bar_interface",
            "network": "bar",
            "description": "Bar network",
            "required": False,
        }]
        result = module._run(interfaces, [])
        expected = {
            "changed": False,
            "ansible_facts": {
                "kolla_foo_interface": "eth0",
                "kolla_bar_interface": "eth1",
            },
            "_ansible_facts_cacheable": False,
        }
        self.assertEqual(expected, result)

    def test__run_one_with_dashes(self):
        variables = copy.deepcopy(self.variables)
        variables["foo_interface"] = "eth-0"
        module = self._create_module(variables)
        interfaces = [{
            "var_name": "kolla_foo_interface",
            "network": "foo",
            "description": "Foo network",
            "required": False,
        }]
        result = module._run(interfaces, [])
        expected = {
            "changed": False,
            "ansible_facts": {
                "kolla_foo_interface": "eth_0",
            },
            "_ansible_facts_cacheable": False,
        }
        self.assertEqual(expected, result)

    def test__run_interface_not_mapped(self):
        module = self._create_module()
        interfaces = [{
            "var_name": "kolla_baz_interface",
            "network": "baz",
            "description": "Baz network",
            "required": True,
        }]
        result = module._run(interfaces, [])
        expected = {
            "changed": False,
            "failed": True,
            "msg": ("Required network 'baz' (Baz network) is not mapped to "
                    "this host"),
        }
        self.assertEqual(expected, result)

    def test__run_interface_not_mapped_not_required(self):
        module = self._create_module()
        interfaces = [{
            "var_name": "kolla_baz_interface",
            "network": "baz",
            "description": "Baz network",
            "required": False,
        }]
        result = module._run(interfaces, [])
        expected = {
            "changed": False,
            "ansible_facts": {},
            "_ansible_facts_cacheable": False,
        }
        self.assertEqual(expected, result)

    def test__run_interface_no_interface(self):
        variables = copy.deepcopy(self.variables)
        del variables["bar_interface"]
        module = self._create_module(variables)
        interfaces = [{
            "var_name": "kolla_bar_interface",
            "network": "bar",
            "description": "Bar network",
            "required": True,
        }]
        result = module._run(interfaces, [])
        expected = {
            "changed": False,
            "failed": True,
            "msg": ("Required network 'bar' (Bar network) does not have an "
                    "interface configured for this host"),
        }
        self.assertEqual(expected, result)

    def test__run_interface_no_interface_not_required(self):
        variables = copy.deepcopy(self.variables)
        del variables["bar_interface"]
        module = self._create_module(variables)
        interfaces = [{
            "var_name": "kolla_bar_interface",
            "network": "bar",
            "description": "Bar network",
            "required": False,
        }]
        result = module._run(interfaces, [])
        expected = {
            "changed": False,
            "ansible_facts": {},
            "_ansible_facts_cacheable": False,
        }
        self.assertEqual(expected, result)

    def test__run_interface_no_interface_not_mapped(self):
        variables = copy.deepcopy(self.variables)
        del variables["bar_interface"]
        module = self._create_module(variables)
        interfaces = [{
            "var_name": "kolla_bar_interface",
            "network": "bar",
            "description": "Bar network",
            "required": True,
        }, {
            "var_name": "kolla_baz_interface",
            "network": "baz",
            "description": "Baz network",
            "required": True,
        }]
        result = module._run(interfaces, [])
        expected = {
            "changed": False,
            "failed": True,
            "msg": ("Required network 'bar' (Bar network) does not have an "
                    "interface configured for this host; Required network "
                    "'baz' (Baz network) is not mapped to this host"),
        }
        self.assertEqual(expected, result)

    def test_run_external_networks_one(self):
        module = self._create_module()
        external_networks = [{
            "network": "foo",
            "required": False,
        }]
        result = module._run([], external_networks)
        expected = {
            "changed": False,
            "ansible_facts": {
                "kolla_neutron_bridge_names": "eth0-ovs",
                "kolla_neutron_external_interfaces": "eth0",
            },
            "_ansible_facts_cacheable": False,
        }
        self.assertEqual(expected, result)

    def test_run_external_networks_two(self):
        module = self._create_module()
        external_networks = [{
            "network": "foo",
            "required": False,
        }, {
            "network": "bar",
            "required": False,
        }]
        result = module._run([], external_networks)
        expected = {
            "changed": False,
            "ansible_facts": {
                "kolla_neutron_bridge_names": "eth0-ovs,eth1-ovs",
                "kolla_neutron_external_interfaces": "eth0,eth1",
            },
            "_ansible_facts_cacheable": False,
        }
        self.assertEqual(expected, result)

    def test_run_external_networks_two_same_interface(self):
        variables = copy.deepcopy(self.variables)
        variables["bar_interface"] = "eth0"
        module = self._create_module(variables)
        external_networks = [{
            "network": "foo",
            "required": False,
        }, {
            "network": "bar",
            "required": False,
        }]
        result = module._run([], external_networks)
        expected = {
            "changed": False,
            "ansible_facts": {
                "kolla_neutron_bridge_names": "eth0-ovs",
                "kolla_neutron_external_interfaces": "eth0",
            },
            "_ansible_facts_cacheable": False,
        }
        self.assertEqual(expected, result)

    def test_run_external_networks_two_vlans(self):
        variables = copy.deepcopy(self.variables)
        variables["foo_interface"] = "eth0.1"
        variables["bar_interface"] = "eth0.2"
        module = self._create_module(variables)
        external_networks = [{
            "network": "foo",
            "required": False,
        }, {
            "network": "bar",
            "required": False,
        }]
        result = module._run([], external_networks)
        expected = {
            "changed": False,
            "ansible_facts": {
                "kolla_neutron_bridge_names": "eth0-ovs",
                "kolla_neutron_external_interfaces": "eth0",
            },
            "_ansible_facts_cacheable": False,
        }
        self.assertEqual(expected, result)

    def test_run_external_networks_bridge(self):
        variables = copy.deepcopy(self.variables)
        variables["foo_interface"] = "breth0"
        module = self._create_module(variables)
        external_networks = [{
            "network": "foo",
            "required": False,
        }]
        result = module._run([], external_networks)
        expected = {
            "changed": False,
            "ansible_facts": {
                "kolla_neutron_bridge_names": "breth0-ovs",
                "kolla_neutron_external_interfaces": "p-breth0-ovs",
            },
            "_ansible_facts_cacheable": False,
        }
        self.assertEqual(expected, result)

    def test_run_external_networks_bridge_vlan(self):
        variables = copy.deepcopy(self.variables)
        variables["foo_interface"] = "breth0.1"
        variables["bar_interface"] = "breth0"
        module = self._create_module(variables)
        external_networks = [{
            "network": "foo",
            "required": False,
        }]
        result = module._run([], external_networks)
        expected = {
            "changed": False,
            "ansible_facts": {
                "kolla_neutron_bridge_names": "breth0-ovs",
                "kolla_neutron_external_interfaces": "p-breth0-ovs",
            },
            "_ansible_facts_cacheable": False,
        }
        self.assertEqual(expected, result)

    def test_run_external_networks_not_mapped(self):
        module = self._create_module()
        external_networks = [{
            "network": "baz",
            "required": True,
        }]
        result = module._run([], external_networks)
        expected = {
            "changed": False,
            "failed": True,
            "msg": ("Required external network 'baz' is not mapped to "
                    "this host"),
        }
        self.assertEqual(expected, result)

    def test_run_external_networks_not_mapped_not_required(self):
        module = self._create_module()
        external_networks = [{
            "network": "baz",
            "required": False,
        }]
        result = module._run([], external_networks)
        expected = {
            "changed": False,
            "ansible_facts": {},
            "_ansible_facts_cacheable": False,
        }
        self.assertEqual(expected, result)

    def test_run_external_networks_no_interface(self):
        variables = copy.deepcopy(self.variables)
        del variables["bar_interface"]
        module = self._create_module(variables)
        external_networks = [{
            "network": "bar",
            "required": True,
        }]
        result = module._run([], external_networks)
        expected = {
            "changed": False,
            "failed": True,
            "msg": ("Required external network 'bar' does not have an "
                    "interface configured for this host"),
        }
        self.assertEqual(expected, result)

    def test_run_external_networks_no_interface_not_required(self):
        variables = copy.deepcopy(self.variables)
        del variables["bar_interface"]
        module = self._create_module(variables)
        external_networks = [{
            "network": "bar",
            "required": False,
        }]
        result = module._run([], external_networks)
        expected = {
            "changed": False,
            "ansible_facts": {},
            "_ansible_facts_cacheable": False,
        }
        self.assertEqual(expected, result)

    def test_run_external_networks_not_mapped_no_interface(self):
        variables = copy.deepcopy(self.variables)
        del variables["bar_interface"]
        module = self._create_module(variables)
        external_networks = [{
            "network": "baz",
            "required": True,
        }, {
            "network": "bar",
            "required": True,
        }]
        result = module._run([], external_networks)
        expected = {
            "changed": False,
            "failed": True,
            "msg": ("Required external network 'baz' is not mapped to "
                    "this host; Required external network 'bar' does not "
                    "have an interface configured for this host"),
        }
        self.assertEqual(expected, result)



# Copyright (c) 2017 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import base64
import glob
import itertools
import logging
import os
import shutil
import subprocess
import sys

import yaml

from kayobe import exception


LOG = logging.getLogger(__name__)


def get_data_files_path(*relative_path):
    """Given a relative path to a data file, return the absolute path"""
    # Detect editable pip install / python setup.py develop and use a path
    # relative to the source directory
    return os.path.join(_get_base_path(), *relative_path)


def _detect_install_prefix(path):
    script_path = os.path.realpath(path)
    script_path = os.path.normpath(script_path)
    components = script_path.split(os.sep)
    # use heuristic: anything before 'lib' in path is the prefix
    if 'lib' not in components:
        return None
    prefix = itertools.takewhile(
        lambda x: x != "lib",
        components
    )
    prefix_path = os.sep.join(prefix)
    return prefix_path


def _get_base_path():
    override = os.environ.get("KAYOBE_DATA_FILES_PATH")
    if override:
        return os.path.join(override)
    egg_glob = os.path.join(
        sys.prefix, 'lib*', 'python*', '*-packages', 'kayobe.egg-link'
    )
    egg_link = glob.glob(egg_glob)
    if egg_link:
        with open(egg_link[0], "r") as f:
            realpath = f.readline().strip()
        return os.path.join(realpath)

    prefix = _detect_install_prefix(__file__)
    if prefix:
        return os.path.join(prefix, "share", "kayobe")

    # Assume uninstalled
    return os.path.join(os.path.realpath(__file__), "..")


def galaxy_role_install(role_file, roles_path, force=False):
    """Install Ansible roles via Ansible Galaxy."""
    cmd = ["ansible-galaxy", "role", "install"]
    cmd += ["--roles-path", roles_path]
    cmd += ["--role-file", role_file]
    if force:
        cmd += ["--force"]
    try:
        run_command(cmd)
    except subprocess.CalledProcessError as e:
        LOG.error("Failed to install Ansible roles from %s via Ansible "
                  "Galaxy: returncode %d", role_file, e.returncode)
        sys.exit(e.returncode)


def galaxy_collection_install(requirements_file, collections_path,
                              force=False):
    requirements = read_yaml_file(requirements_file)
    if not isinstance(requirements, dict):
        # Handle legacy role list format, which causes the command to fail.
        return
    cmd = ["ansible-galaxy", "collection", "install"]
    cmd += ["--collections-path", collections_path]
    cmd += ["--requirements-file", requirements_file]
    if force:
        cmd += ["--force"]
    try:
        run_command(cmd)
    except subprocess.CalledProcessError as e:
        LOG.error("Failed to install Ansible collections from %s via Ansible "
                  "Galaxy: returncode %d", requirements_file, e.returncode)
        sys.exit(e.returncode)


def galaxy_remove(roles_to_remove, roles_path):

    """Remove Ansible roles via Ansible Galaxy."""
    cmd = ["ansible-galaxy", "role", "remove"]
    cmd += ["--roles-path", roles_path]
    cmd += roles_to_remove
    try:
        run_command(cmd)
    except subprocess.CalledProcessError as e:
        LOG.error("Failed to remove Ansible roles %s via Ansible "
                  "Galaxy: returncode %d",
                  ",".join(roles_to_remove), e.returncode)
        sys.exit(e.returncode)


def read_file(path, mode="r"):
    """Read the content of a file."""
    with open(path, mode) as f:
        return f.read()


def read_yaml_file(path):
    """Read and decode a YAML file."""
    try:
        content = read_file(path)
    except IOError as e:
        print("Failed to open config dump file %s: %s" %
              (path, repr(e)))
        sys.exit(1)
    try:
        return yaml.safe_load(content)
    except yaml.YAMLError as e:
        print("Failed to decode config dump YAML file %s: %s" %
              (path, repr(e)))
        sys.exit(1)


def is_readable_dir(path):
    """Check whether a path references a readable directory."""
    if not os.path.exists(path):
        return {"result": False, "message": "Path does not exist"}
    if not os.path.isdir(path):
        return {"result": False, "message": "Path is not a directory"}
    if not os.access(path, os.R_OK):
        return {"result": False, "message": "Directory is not readable"}
    return {"result": True}


def is_readable_file(path):
    """Check whether a path references a readable file."""
    if not os.path.exists(path):
        return {"result": False, "message": "Path does not exist"}
    if not os.path.isfile(path):
        return {"result": False, "message": "Path is not a file"}
    if not os.access(path, os.R_OK):
        return {"result": False, "message": "File is not readable"}
    return {"result": True}


def run_command(cmd, quiet=False, check_output=False, **kwargs):
    """Run a command, checking the output.

    :param quiet: Redirect output to /dev/null
    :param check_output: Whether to return the output of the command
    :returns: The output of the command if check_output is true
    """
    if isinstance(cmd, str):
        cmd_string = cmd
    else:
        cmd_string = " ".join(cmd)
    LOG.debug("Running command: %s", cmd_string)
    if quiet:
        with open("/dev/null", "w") as devnull:
            kwargs["stdout"] = devnull
            kwargs["stderr"] = devnull
            subprocess.check_call(cmd, **kwargs)
    elif check_output:
        return subprocess.check_output(cmd, **kwargs)
    else:
        subprocess.check_call(cmd, **kwargs)


def quote_and_escape(value):
    """Quote and escape a string.

    Adds enclosing single quotes to the string passed, and escapes single
    quotes within the string using backslashes. This is useful for passing
    'extra vars' to Ansible. Without this, Ansible only uses the part of the
    string up to the first whitespace.

    :param value: the string to quote and escape.
    :returns: the quoted and escaped string.
    """
    if not isinstance(value, str):
        return value
    return "'" + value.replace("'", "'\\''") + "'"


def escape_jinja(string):
    """Escapes a string so that jinja template variables are not expanded

    :param string: the string to escape
    :return: the escaped string
    """
    # We base64 encode the string to avoid the need to escape characters.
    # This is because ansible has some parsing quirks that makes it fairly
    # hard to escape stuff in generic way.
    # See: https://github.com/ansible/ansible/issues/10464

    b64_value = base64.b64encode(string.encode())
    return ''.join(('{{', "'", b64_value.decode(), "' | b64decode ", '}}'))


def intersect_limits(args_limit, cli_limit):
    """Create an Ansible host pattern of the intersection of two patterns.

    :param args_limit: user-specified limit, or None.
    :param cli_limit: limit originating from this CLI, or None.
    :returns: a string representing an intersection of the two patterns.
    """
    # NOTE(mgoddard): Ansible uses either commas (,) or colons (:) to separate
    # parts of a host pattern. An intersection is specified using a separator
    # followed by an ampersand (&). If a mix of comma and colon separators is
    # used, Ansible picks one and treats the other as part of the host pattern.
    # This leads to hard to diagnose errors. Try to determine which separator
    # the user has specified, and be consistent. Error if both are used.
    if args_limit and ',' in args_limit:
        if ':' in args_limit:
            raise exception.Error("Invalid format for host limit argument. "
                                  "Cannot mix commas and colons to separate "
                                  "hosts")
        separator = ',&'
    else:
        separator = ':&'
    limits = [l for l in [args_limit, cli_limit] if l]
    return separator.join(limits)


def copy_dir(src, dest, exclude=None):
    """Copy recursively a directory.

    :param src: path of the source directory
    :param dest: destination path, will be created if it does not exist
    :param exclude: names of files or directories at the root of the source
                    directory to exclude during copy
    """
    if exclude is None:
        exclude = []

    if not os.path.isdir(dest):
        os.mkdir(dest)

    for file in os.listdir(src):
        if file in exclude:
            continue

        src_path = os.path.join(src, file)
        dest_path = os.path.join(dest, file)
        if os.path.isdir(src_path):
            copy_dir(src_path, dest_path)
        else:
            shutil.copy2(src_path, dest_path)



# Copyright (c) 2018 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import configparser
from io import StringIO


def test_file(host, path, owner='root', group='root'):
    """Test an expected file.

    Validate that the file exists and has the correct ownership.
    """
    f = host.file(path)

    assert f.exists
    assert f.is_file
    assert owner == f.user
    assert group == f.group


def test_ini_file(host, path, owner='root', group='root', expected=None):
    """Test an expected INI file.

    Validate that the file exists, has the correct ownership, format and
    expected contents.

    :param expected: a dict of dicts providing config that should be present.
    """
    test_file(host, path, owner, group)

    sio = StringIO(host.file(path).content_string)
    parser = configparser.RawConfigParser()
    parser.read_file(sio)

    if expected is None:
        return

    for exp_section_name, exp_section in expected.items():
        assert parser.has_section(exp_section_name)
        for exp_key, exp_value in exp_section.items():
            assert parser.has_option(exp_section_name, exp_key)
            assert parser.get(exp_section_name, exp_key) == exp_value


def test_directory(host, path, owner='root', group='root'):
    """Test an expected directory.

    Validate that the directory exists and has the correct ownership.
    """
    d = host.file(path)

    assert d.exists
    assert d.is_directory
    assert owner == d.user
    assert group == d.group


def test_path_absent(host, path):
    """Test a path expected to not exist."""
    p = host.file(path)

    assert not p.exists



# Copyright (c) 2017 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import os
import subprocess
import unittest
from unittest import mock

from kayobe import ansible
from kayobe import kolla_ansible
from kayobe import utils
from kayobe import vault


@mock.patch.object(os, "getcwd", new=lambda: "/path/to/cwd")
@mock.patch.dict(os.environ, clear=True)
class TestCase(unittest.TestCase):

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(kolla_ansible, "_validate_args")
    def test_run(self, mock_validate, mock_run):
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        kolla_ansible.add_args(parser)
        vault.add_args(parser)
        parsed_args = parser.parse_args([])
        kolla_ansible.run(parsed_args, "command", "overcloud")
        expected_cmd = [
            ".", "/path/to/cwd/venvs/kolla-ansible/bin/activate", "&&",
            "kolla-ansible", "command",
            "--inventory", "/etc/kolla/inventory/overcloud",
        ]
        expected_cmd = " ".join(expected_cmd)
        mock_run.assert_called_once_with(expected_cmd, shell=True, quiet=False,
                                         env={})

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(kolla_ansible, "_validate_args")
    def test_run_all_the_args(self, mock_validate, mock_run):
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        kolla_ansible.add_args(parser)
        vault.add_args(parser)
        args = [
            "-C",
            "-D",
            "--kolla-config-path", "/path/to/config",
            "-ke", "ev_name1=ev_value1",
            "-ki", "/path/to/inventory",
            "-kl", "host1:host2",
            "-kt", "tag1,tag2",
        ]
        parsed_args = parser.parse_args(args)
        kolla_ansible.run(parsed_args, "command", "overcloud")
        expected_cmd = [
            ".", "/path/to/cwd/venvs/kolla-ansible/bin/activate", "&&",
            "kolla-ansible", "command",
            "--inventory", "/path/to/inventory",
            "--configdir", "/path/to/config",
            "--passwords", "/path/to/config/passwords.yml",
            "-e", "ev_name1=ev_value1",
            "--limit", "'host1:host2'",
            "--tags", "tag1,tag2",
        ]
        expected_cmd = " ".join(expected_cmd)
        expected_env = {"EXTRA_OPTS": " --check --diff"}
        mock_run.assert_called_once_with(expected_cmd, shell=True, quiet=False,
                                         env=expected_env)

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(kolla_ansible, "_validate_args")
    @mock.patch.object(vault, "_ask_vault_pass")
    def test_run_all_the_long_args(self, mock_ask, mock_validate, mock_run):
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        kolla_ansible.add_args(parser)
        vault.add_args(parser)
        mock_ask.return_value = "test-pass"
        args = [
            "--ask-vault-pass",
            "--check",
            "--diff",
            "--kolla-config-path", "/path/to/config",
            "--kolla-extra-vars", "ev_name1=ev_value1",
            "--kolla-inventory", "/path/to/inventory",
            "--kolla-limit", "host1:host2",
            "--kolla-skip-tags", "tag3,tag4",
            "--kolla-tags", "tag1,tag2",
        ]
        parsed_args = parser.parse_args(args)
        mock_run.return_value = "/path/to/kayobe-vault-password-helper"
        kolla_ansible.run(parsed_args, "command", "overcloud")
        expected_cmd = [
            ".", "/path/to/cwd/venvs/kolla-ansible/bin/activate", "&&",
            "kolla-ansible", "command",
            "--key", "/path/to/kayobe-vault-password-helper",
            "--inventory", "/path/to/inventory",
            "--configdir", "/path/to/config",
            "--passwords", "/path/to/config/passwords.yml",
            "-e", "ev_name1=ev_value1",
            "--limit", "'host1:host2'",
            "--skip-tags", "tag3,tag4",
            "--tags", "tag1,tag2",
        ]
        expected_cmd = " ".join(expected_cmd)
        expected_env = {"EXTRA_OPTS": " --check --diff",
                        "KAYOBE_VAULT_PASSWORD": "test-pass"}
        expected_calls = [
            mock.call(["which", "kayobe-vault-password-helper"],
                      check_output=True, universal_newlines=True),
            mock.call(expected_cmd, shell=True, quiet=False, env=expected_env)
        ]
        self.assertEqual(expected_calls, mock_run.mock_calls)

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(kolla_ansible, "_validate_args")
    @mock.patch.object(vault, "update_environment")
    def test_run_vault_password_file(self, mock_update, mock_validate,
                                     mock_run):
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        kolla_ansible.add_args(parser)
        vault.add_args(parser)
        args = [
            "--vault-password-file", "/path/to/vault/pw",
        ]
        parsed_args = parser.parse_args(args)
        kolla_ansible.run(parsed_args, "command", "overcloud")
        expected_cmd = [
            ".", "/path/to/cwd/venvs/kolla-ansible/bin/activate", "&&",
            "kolla-ansible", "command",
            "--key", "/path/to/vault/pw",
            "--inventory", "/etc/kolla/inventory/overcloud",
        ]
        expected_cmd = " ".join(expected_cmd)
        mock_run.assert_called_once_with(expected_cmd, shell=True, quiet=False,
                                         env={})
        mock_update.assert_called_once_with(mock.ANY, {})

    @mock.patch.dict(os.environ, {"KAYOBE_VAULT_PASSWORD": "test-pass"})
    @mock.patch.object(utils, "run_command")
    @mock.patch.object(kolla_ansible, "_validate_args")
    @mock.patch.object(vault, "update_environment")
    def test_run_vault_password_helper(self, mock_update, mock_vars, mock_run):
        mock_vars.return_value = []
        parser = argparse.ArgumentParser()
        mock_run.return_value = "/path/to/kayobe-vault-password-helper"
        ansible.add_args(parser)
        kolla_ansible.add_args(parser)
        vault.add_args(parser)
        mock_run.assert_called_once_with(
            ["which", "kayobe-vault-password-helper"], check_output=True,
            universal_newlines=True)
        mock_run.reset_mock()
        parsed_args = parser.parse_args([])
        kolla_ansible.run(parsed_args, "command", "overcloud")
        expected_cmd = [
            ".", "/path/to/cwd/venvs/kolla-ansible/bin/activate", "&&",
            "kolla-ansible", "command",
            "--key", "/path/to/kayobe-vault-password-helper",
            "--inventory", "/etc/kolla/inventory/overcloud",
        ]
        expected_cmd = " ".join(expected_cmd)
        expected_env = {"KAYOBE_VAULT_PASSWORD": "test-pass"}
        mock_run.assert_called_once_with(expected_cmd, shell=True, quiet=False,
                                         env=expected_env)
        mock_update.assert_called_once_with(mock.ANY, expected_env)

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(kolla_ansible, "_validate_args")
    def test_run_func_args(self, mock_validate, mock_run):
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        kolla_ansible.add_args(parser)
        vault.add_args(parser)
        args = [
            "--kolla-extra-vars", "ev_name1=ev_value1",
            "--kolla-tags", "tag1,tag2",
        ]
        parsed_args = parser.parse_args(args)
        kwargs = {
            "extra_vars": {"ev_name2": "ev_value2"},
            "tags": "tag3,tag4",
            "verbose_level": 1,
            "extra_args": ["--arg1", "--arg2"],
        }
        kolla_ansible.run(parsed_args, "command", "overcloud", **kwargs)
        expected_cmd = [
            ".", "/path/to/cwd/venvs/kolla-ansible/bin/activate", "&&",
            "kolla-ansible", "command",
            "-v",
            "--inventory", "/etc/kolla/inventory/overcloud",
            "-e", "ev_name1=ev_value1",
            "-e", "ev_name2='ev_value2'",
            "--tags", "tag1,tag2,tag3,tag4",
            "--arg1", "--arg2",
        ]
        expected_cmd = " ".join(expected_cmd)
        mock_run.assert_called_once_with(expected_cmd, shell=True, quiet=False,
                                         env={})

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(utils, "is_readable_file")
    @mock.patch.object(kolla_ansible, "_validate_args")
    def test_run_custom_ansible_cfg(self, mock_validate, mock_readable,
                                    mock_run):
        mock_readable.return_value = {"result": True}
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        kolla_ansible.add_args(parser)
        vault.add_args(parser)
        parsed_args = parser.parse_args([])
        kolla_ansible.run(parsed_args, "command", "overcloud")
        expected_cmd = [
            ".", "/path/to/cwd/venvs/kolla-ansible/bin/activate", "&&",
            "kolla-ansible", "command",
            "--inventory", "/etc/kolla/inventory/overcloud",
        ]
        expected_cmd = " ".join(expected_cmd)
        expected_env = {"ANSIBLE_CONFIG": "/etc/kayobe/kolla/ansible.cfg"}
        mock_run.assert_called_once_with(expected_cmd, shell=True, quiet=False,
                                         env=expected_env)
        mock_readable.assert_called_once_with("/etc/kayobe/kolla/ansible.cfg")

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(utils, "is_readable_file")
    @mock.patch.object(kolla_ansible, "_validate_args")
    def test_run_custom_ansible_cfg_2(self, mock_validate, mock_readable,
                                      mock_run):
        mock_readable.side_effect = [{"result": False}, {"result": True}]
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        kolla_ansible.add_args(parser)
        vault.add_args(parser)
        parsed_args = parser.parse_args([])
        kolla_ansible.run(parsed_args, "command", "overcloud")
        expected_cmd = [
            ".", "/path/to/cwd/venvs/kolla-ansible/bin/activate", "&&",
            "kolla-ansible", "command",
            "--inventory", "/etc/kolla/inventory/overcloud",
        ]
        expected_cmd = " ".join(expected_cmd)
        expected_env = {"ANSIBLE_CONFIG": "/etc/kayobe/ansible.cfg"}
        mock_run.assert_called_once_with(expected_cmd, shell=True, quiet=False,
                                         env=expected_env)
        expected_calls = [
            mock.call("/etc/kayobe/kolla/ansible.cfg"),
            mock.call("/etc/kayobe/ansible.cfg"),
        ]
        self.assertEqual(mock_readable.call_args_list, expected_calls)

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(utils, "is_readable_file")
    @mock.patch.object(kolla_ansible, "_validate_args")
    def test_run_custom_ansible_cfg_env(self, mock_validate, mock_readable,
                                        mock_run):
        mock_readable.return_value = {"result": True}
        os.environ["ANSIBLE_CONFIG"] = "/path/to/ansible.cfg"
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        kolla_ansible.add_args(parser)
        vault.add_args(parser)
        parsed_args = parser.parse_args([])
        kolla_ansible.run(parsed_args, "command", "overcloud")
        expected_cmd = [
            ".", "/path/to/cwd/venvs/kolla-ansible/bin/activate", "&&",
            "kolla-ansible", "command",
            "--inventory", "/etc/kolla/inventory/overcloud",
        ]
        expected_cmd = " ".join(expected_cmd)
        expected_env = {"ANSIBLE_CONFIG": "/path/to/ansible.cfg"}
        mock_run.assert_called_once_with(expected_cmd, shell=True, quiet=False,
                                         env=expected_env)
        mock_readable.assert_called_once_with("/etc/kayobe/kolla/ansible.cfg")

    @mock.patch.object(utils, "run_command")
    @mock.patch.object(kolla_ansible, "_validate_args")
    def test_run_failure(self, mock_validate, mock_run):
        parser = argparse.ArgumentParser()
        ansible.add_args(parser)
        kolla_ansible.add_args(parser)
        vault.add_args(parser)
        parsed_args = parser.parse_args([])
        mock_run.side_effect = subprocess.CalledProcessError(1, "dummy")
        self.assertRaises(SystemExit,
                          kolla_ansible.run, parsed_args, "command",
                          "overcloud")



#!/usr/bin/python3

# Copyright (c) 2017 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# TODO(wszumski): If we have multiple conductors and they are on different machines
# we could make a pool per machine.

DOCUMENTATION = """
module: console_allocation
short_description: Allocate a serial console TCP port for an Ironic node from a pool
author: Mark Goddard (mark@stackhpc.com) and Will Szumski (will@stackhpc.com)
options:
  - option-name: nodes
    description: List of Names or UUIDs corresponding to Ironic Nodes
    required: True
    type: list
  - option-name: allocation_pool_start
    description: First address of the pool from which to allocate
    required: True
    type: int
  - option-name: allocation_pool_end
    description: Last address of the pool from which to allocate
    required: True
    type: int
  - option-name: allocation_file
    description: >
      Path to a file in which to store the allocations. Will be created if it
      does not exist.
    required: True
    type: string
requirements:
  - PyYAML
"""

EXAMPLES = """
- name: Ensure Ironic node has a TCP port assigned for it's serial console
  console_allocation:
    nodes: ['node-1', 'node-2']
    allocation_pool_start: 30000
    allocation_pool_end: 31000
    allocation_file: /path/to/allocation/file.yml
"""

RETURN = """
ports:
  description: >
    A dictionary mapping the node name to the allocated serial console TCP port
  returned: success
  type: dict
  sample: { 'node1' : 30000, 'node2':300001 }
"""

from ansible.module_utils.basic import *
import sys

# Store a list of import errors to report to the user.
IMPORT_ERRORS=[]
try:
    import yaml
except Exception as e:
    IMPORT_ERRORS.append(e)


def read_allocations(module):
    """Read TCP port allocations from the allocation file."""
    filename = module.params['allocation_file']
    try:
        with open(filename, 'r') as f:
            content = yaml.safe_load(f)
    except IOError as e:
        if e.errno == errno.ENOENT:
            # Ignore ENOENT - we will create the file.
            return {}
        module.fail_json(msg="Failed to open allocation file %s for reading" % filename)
    except yaml.YAMLError as e:
        module.fail_json(msg="Failed to parse allocation file %s as YAML" % filename)
    if content is None:
        # If the file is empty, yaml.safe_load() will return None.
        content = {}
    return content


def write_allocations(module, allocations):
    """Write TCP port allocations to the allocation file."""
    filename = module.params['allocation_file']
    try:
        with open(filename, 'w') as f:
            yaml.dump(allocations, f, default_flow_style=False)
    except IOError as e:
        module.fail_json(msg="Failed to open allocation file %s for writing" % filename)
    except yaml.YAMLError as e:
        module.fail_json(msg="Failed to dump allocation file %s as YAML" % filename)

def is_valid_port(port):
    try:
        int(port)
    except ValueError:
        return False
    if port < 0:
        return False
    if port > 65535:
        return False
    return True


def update_allocation(module, allocations):
    """Allocate a TCP port of an Ironic serial console.

    :param module: AnsibleModule instance
    :param allocations: Existing IP address allocations
    """
    nodes = module.params['nodes']

    allocation_pool_start = module.params['allocation_pool_start']
    allocation_pool_end = module.params['allocation_pool_end']
    result = {
        'changed': False,
        'ports': {}
    }
    object_name = "serial_console_allocations"
    console_allocations = allocations.setdefault(object_name, {})
    invalid_allocations = {node: port for node, port in console_allocations.items()
                           if not is_valid_port(port)}
    if invalid_allocations:
        module.fail_json(msg="Found invalid existing allocations in %s: %s" %
            (object_name,
             ", ".join("%s: %s" % (node, port)
                       for node, port in invalid_allocations.items())))

    allocated_consoles = { int(x) for x in  console_allocations.values() }
    allocation_pool = { x for x in range(allocation_pool_start, allocation_pool_end + 1) }
    free_ports = list(allocation_pool - allocated_consoles)
    free_ports.sort(reverse=True)

    for node in nodes:
        if node not in console_allocations:
            if len(free_ports) < 1:
                module.fail_json(msg="No unallocated TCP ports for %s in %s" % (node, object_name))
            result['changed'] = True
            free_port = free_ports.pop()
            console_allocations[node] = free_port
        result['ports'][node] = console_allocations[node]
    return result


def allocate(module):
    """Allocate a TCP port an ironic serial console, updating the allocation file."""
    allocations = read_allocations(module)
    result = update_allocation(module, allocations)
    if result['changed'] and not module.check_mode:
        write_allocations(module, allocations)
    return result


def main():
    module = AnsibleModule(
        argument_spec=dict(
            nodes=dict(required=True, type='list'),
            allocation_pool_start=dict(required=True, type='int'),
            allocation_pool_end=dict(required=True, type='int'),
            allocation_file=dict(required=True, type='str'),
        ),
        supports_check_mode=True,
    )

    # Fail if there were any exceptions when importing modules.
    if IMPORT_ERRORS:
        module.fail_json(msg="Import errors: %s" %
                         ", ".join([repr(e) for e in IMPORT_ERRORS]))

    try:
        results = allocate(module)
    except Exception as e:
        module.fail_json(msg="Failed to allocate TCP port: %s" % repr(e))
    else:
        module.exit_json(**results)


if __name__ == '__main__':
    main()



# Copyright (c) 2020 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ansible.plugins.action import ActionBase


class ConfigError(Exception):
    pass


class ActionModule(ActionBase):
    """Kolla Ansible host vars action plugin

    This class provides an Ansible action module that returns facts
    representing host variables to be passed to Kolla Ansible.
    """

    TRANSFERS_FILES = False

    def run(self, tmp=None, task_vars=None):
        if task_vars is None:
            task_vars = dict()

        result = super(ActionModule, self).run(tmp, task_vars)
        del tmp  # tmp no longer has any effect

        # Module arguments:
        # interfaces: a list of dicts, each containing 'network', 'required',
        #             and 'description' keys. Each describes a Kolla Ansible
        #             interface variable.
        # external_networks: a list of dicts, each containing 'network', and
        #                    'required' keys. Each describes an external
        #                    network.
        interfaces = self._task.args["interfaces"]
        external_networks = self._task.args.get('external_networks', [])
        result.update(self._run(interfaces, external_networks))
        return result

    def _run(self, interfaces, external_networks):
        result = {}
        facts = {}
        errors = []

        # Kolla Ansible interface facts.
        for interface in interfaces:
            try:
                iface = self._get_interface_fact(interface["network"],
                                                 interface["required"],
                                                 interface["description"])
                if iface:
                    facts[interface["var_name"]] = iface
            except ConfigError as e:
                errors.append(str(e))

        # Build a list of external network interfaces.
        external_interfaces = []
        for network in external_networks:
            try:
                iface = self._get_external_interface(network["network"],
                                                     network["required"])
                if iface and iface not in external_interfaces:
                    external_interfaces.append(iface)
            except ConfigError as e:
                errors.append(str(e))

        if external_interfaces:
            facts.update(self._get_external_interface_facts(
                external_interfaces))

        result['changed'] = False
        if errors:
            result['failed'] = True
            result['msg'] = "; ".join(errors)
        else:
            result['ansible_facts'] = facts
            result['_ansible_facts_cacheable'] = False
        return result

    def _get_interface_fact(self, net_name, required, description):
        # Check whether the network is mapped to this host.
        condition = "{{ '%s' in network_interfaces }}" % net_name
        condition = self._templar.template(condition)
        if condition:
            # Get the network interface for this network.
            iface = ("{{ '%s' | net_interface }}" % net_name)
            iface = self._templar.template(iface)
            if iface:
                # Ansible fact names replace dashes with underscores.
                # FIXME(mgoddard): Is this still required?
                iface = iface.replace('-', '_')
            if required and not iface:
                msg = ("Required network '%s' (%s) does not have an interface "
                       "configured for this host" % (net_name, description))
                raise ConfigError(msg)
            return iface
        elif required:
            msg = ("Required network '%s' (%s) is not mapped to this host" %
                   (net_name, description))
            raise ConfigError(msg)

    def _get_external_interface(self, net_name, required):
        condition = "{{ '%s' in network_interfaces }}" % net_name
        condition = self._templar.template(condition)
        if condition:
            iface = self._templar.template("{{ '%s' | net_interface }}" %
                                           net_name)
            if iface:
                # When these networks are VLANs, we need to use the
                # underlying tagged bridge interface rather than the
                # untagged interface. We therefore strip the .<vlan> suffix
                # of the interface name. We use a union here as a single
                # tagged interface may be shared between these networks.
                vlan = self._templar.template("{{ '%s' | net_vlan }}" %
                                              net_name)
                if vlan and iface.endswith(".%s" % vlan):
                    iface = iface.replace(".%s" % vlan, "")
                return iface
            elif required:
                raise ConfigError("Required external network '%s' does not "
                                  "have an interface configured for this host"
                                  % net_name)
        elif required:
            raise ConfigError("Required external network '%s' is not mapped "
                              "to this host" % net_name)

    def _get_external_interface_facts(self, external_interfaces):
        neutron_bridge_names = []
        neutron_external_interfaces = []
        bridge_suffix = self._templar.template(
            "{{ network_bridge_suffix_ovs }}")
        patch_prefix = self._templar.template("{{ network_patch_prefix }}")
        patch_suffix = self._templar.template("{{ network_patch_suffix_ovs }}")
        for interface in external_interfaces:
            is_bridge = ("{{ '%s' in (network_interfaces |"
                         "net_select_bridges |"
                         "map('net_interface')) }}" % interface)
            is_bridge = self._templar.template(is_bridge)
            neutron_bridge_names.append(interface + bridge_suffix)
            # For a bridge, use a veth pair connected to the bridge. Otherwise
            # use the interface directly.
            if is_bridge:
                external_interface = patch_prefix + interface + patch_suffix
            else:
                external_interface = interface
            neutron_external_interfaces.append(external_interface)
        return {
            "kolla_neutron_bridge_names": ",".join(neutron_bridge_names),
            "kolla_neutron_external_interfaces": ",".join(
                neutron_external_interfaces),
        }



# Copyright (c) 2021 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from ansible import errors
import jinja2
import netaddr
import re

from kayobe.plugins.filter import utils


def get_and_validate_interface(context, name, inventory_hostname):
    """Return a validated interface for a network.

    :param context: a Jinja2 Context object.
    :param name: name of the network.
    :param inventory_hostname: Ansible inventory hostname.
    :returns: a validated interface for a network.
    :raises: ansible.errors.AnsibleFilterError
    """
    device = net_interface(context, name, inventory_hostname)
    if not device:
        raise errors.AnsibleFilterError(
            "Network interface for network '%s' on host '%s' not found" %
            (name, inventory_hostname))
    return device


def _get_veth_interface(context, bridge, inventory_hostname):
    """Return a veth link name for a bridge.

    :param context: a Jinja2 Context object.
    :param bridge: name of the bridge interface into which the veth is plugged.
    :param inventory_hostname: Ansible inventory hostname.
    :returns: a veth link name for a bridge.
    """
    prefix = utils.get_hostvar(context, 'network_patch_prefix',
                               inventory_hostname)
    suffix = utils.get_hostvar(context, 'network_patch_suffix_phy',
                               inventory_hostname)
    return prefix + bridge + suffix


def _get_veth_peer(context, bridge, inventory_hostname):
    """Return a veth peer name for a bridge.

    :param context: a Jinja2 Context object.
    :param bridge: name of the bridge interface into which the veth is plugged.
    :param inventory_hostname: Ansible inventory hostname.
    :returns: a veth peer name for a bridge.
    """
    prefix = utils.get_hostvar(context, 'network_patch_prefix',
                               inventory_hostname)
    suffix = utils.get_hostvar(context, 'network_patch_suffix_ovs',
                               inventory_hostname)
    return prefix + bridge + suffix


def get_ovs_veths(context, names, inventory_hostname):
    """Return a list of dicts describing veth pairs to plug into Open vSwitch.

    :param context: a Jinja2 Context object.
    :param names: list of names of networks.
    :param inventory_hostname: Ansible inventory hostname.
    :returns: a list of dicts describing veth pairs. Each dict has keys 'name',
              'peer', 'bridge', and 'mtu'.
    """
    # The following networks need to be plugged into Open vSwitch:
    # * workload provisioning network
    # * workload cleaning network
    # * neutron external networks
    ironic_networks = [
        utils.get_hostvar(context, 'provision_wl_net_name',
                          inventory_hostname),
        utils.get_hostvar(context, 'cleaning_net_name', inventory_hostname),
    ]
    external_networks = utils.get_hostvar(context, 'external_net_names',
                                          inventory_hostname)
    veth_networks = ironic_networks + (external_networks or [])

    # Make a list of all bridge interfaces.
    bridges = net_select_bridges(context, names, inventory_hostname)
    bridge_interfaces = [net_interface(context, bridge, inventory_hostname)
                         for bridge in bridges]

    # Dict mapping bridge interfaces to the MTU of a connected veth pair.
    veth_mtu_map = {}
    for name in veth_networks:
        if name not in names:
            continue
        device = get_and_validate_interface(context, name, inventory_hostname)
        # When these networks are VLANs, we need to use the underlying tagged
        # interface rather than the untagged interface. We therefore strip the
        # .<vlan> suffix of the interface name. We use a union here as a single
        # tagged interface may be shared between these networks.
        vlan = net_vlan(context, name, inventory_hostname)
        if vlan:
            parent_or_device = get_vlan_parent(device, vlan)
        else:
            parent_or_device = device
        if parent_or_device in bridge_interfaces:
            # Determine the MTU as the maximum of all subinterface MTUs. Only
            # interfaces with an explicit MTU set will be taken account of. If
            # no interface has an explicit MTU set, then the corresponding veth
            # will not either.
            # Allow for the case where an MTU is not specified.
            mtu = net_mtu(context, name, inventory_hostname)
            veth_mtu_map.setdefault(parent_or_device, mtu)
            if (veth_mtu_map.get(parent_or_device) or 0) < (mtu or 0):
                veth_mtu_map[parent_or_device] = mtu

    return [
        {
            'name': _get_veth_interface(context, bridge, inventory_hostname),
            'peer': _get_veth_peer(context, bridge, inventory_hostname),
            'bridge': bridge,
            'mtu': mtu
        }
        for bridge, mtu in veth_mtu_map.items()
    ]


def get_vlan_parent(device, vlan):
    """Return the parent interface of a VLAN subinterface.

    :param device: VLAN interface name.
    :param vlan: VLAN ID.
    :returns: parent interface name.
    """
    return re.sub(r'\.{}$'.format(vlan), '', device)


@jinja2.contextfilter
def net_attr(context, name, attr, inventory_hostname=None):
    var_name = "%s_%s" % (name, attr)
    return utils.get_hostvar(context, var_name, inventory_hostname)


def _make_attr_filter(attr):
    @jinja2.contextfilter
    def func(context, name, inventory_hostname=None):
        return net_attr(context, name, attr, inventory_hostname)
    return func


@jinja2.contextfilter
def net_vip_address(context, name, inventory_hostname=None):
    return net_attr(context, name, 'vip_address', inventory_hostname)


@jinja2.contextfilter
def net_ip(context, name, inventory_hostname=None):
    ips = net_attr(context, name, 'ips', inventory_hostname)
    if ips:
        if inventory_hostname is None:
            inventory_hostname = utils.get_hostvar(context,
                                                   "inventory_hostname")
        return ips.get(inventory_hostname)


@jinja2.contextfilter
def net_interface(context, name, inventory_hostname=None):
    return net_attr(context, name, 'interface', inventory_hostname)


@jinja2.contextfilter
def net_cidr(context, name, inventory_hostname=None):
    return net_attr(context, name, 'cidr', inventory_hostname)


@jinja2.contextfilter
def net_mask(context, name, inventory_hostname=None):
    cidr = net_cidr(context, name, inventory_hostname)
    return str(netaddr.IPNetwork(cidr).netmask) if cidr is not None else None


@jinja2.contextfilter
def net_prefix(context, name, inventory_hostname=None):
    cidr = net_cidr(context, name, inventory_hostname)
    return str(netaddr.IPNetwork(cidr).prefixlen) if cidr is not None else None


@jinja2.contextfilter
def net_gateway(context, name, inventory_hostname=None):
    return net_attr(context, name, 'gateway', inventory_hostname)


@jinja2.contextfilter
def net_allocation_pool_start(context, name, inventory_hostname=None):
    return net_attr(context, name, 'allocation_pool_start', inventory_hostname)


@jinja2.contextfilter
def net_allocation_pool_end(context, name, inventory_hostname=None):
    return net_attr(context, name, 'allocation_pool_end', inventory_hostname)


@jinja2.contextfilter
def net_inspection_allocation_pool_start(context, name,
                                         inventory_hostname=None):
    return net_attr(context, name, 'inspection_allocation_pool_start',
                    inventory_hostname)


@jinja2.contextfilter
def net_inspection_allocation_pool_end(context, name, inventory_hostname=None):
    return net_attr(context, name, 'inspection_allocation_pool_end',
                    inventory_hostname)


net_inspection_gateway = _make_attr_filter('inspection_gateway')


@jinja2.contextfilter
def net_neutron_allocation_pool_start(context, name, inventory_hostname=None):
    return net_attr(context, name, 'neutron_allocation_pool_start',
                    inventory_hostname)


@jinja2.contextfilter
def net_neutron_allocation_pool_end(context, name, inventory_hostname=None):
    return net_attr(context, name, 'neutron_allocation_pool_end',
                    inventory_hostname)


net_neutron_gateway = _make_attr_filter('neutron_gateway')


@jinja2.contextfilter
def net_vlan(context, name, inventory_hostname=None):
    return net_attr(context, name, 'vlan', inventory_hostname)


@jinja2.contextfilter
def net_mtu(context, name, inventory_hostname=None):
    mtu = net_attr(context, name, 'mtu', inventory_hostname)
    if mtu is not None:
        mtu = int(mtu)
    return mtu


net_routes = _make_attr_filter('routes')
net_rules = _make_attr_filter('rules')
net_physical_network = _make_attr_filter('physical_network')
net_bootproto = _make_attr_filter('bootproto')
net_defroute = _make_attr_filter('defroute')
net_ethtool_opts = _make_attr_filter('ethtool_opts')
net_zone = _make_attr_filter('zone')


@jinja2.contextfilter
def net_libvirt_network_name(context, name, inventory_hostname=None):
    """Return the configured Libvirt name for a network.

    If no Libvirt name is configured, the network's name is returned.
    """
    libvirt_name = net_attr(context, name, 'libvirt_network_name',
                            inventory_hostname)
    return libvirt_name or name


@jinja2.contextfilter
def net_bridge_ports(context, name, inventory_hostname=None):
    return net_attr(context, name, 'bridge_ports', inventory_hostname)


net_bond_mode = _make_attr_filter('bond_mode')
net_bond_slaves = _make_attr_filter('bond_slaves')
net_bond_miimon = _make_attr_filter('bond_miimon')
net_bond_updelay = _make_attr_filter('bond_updelay')
net_bond_downdelay = _make_attr_filter('bond_downdelay')
net_bond_xmit_hash_policy = _make_attr_filter('bond_xmit_hash_policy')
net_bond_lacp_rate = _make_attr_filter('bond_lacp_rate')


def _route_obj(route):
    """Return a dict representation of an IP route.

    The returned dict is compatible with the route item of the
    interfaces_ether_interfaces and interfaces_bridge_interfaces variables in
    the MichaelRigart.interfaces role.
    """
    net = netaddr.IPNetwork(route['cidr'])
    route_obj = {
        'network': str(net.network),
        'netmask': str(net.netmask),
    }
    optional = {
        'gateway',
        'table',
        'options',
    }
    for option in optional:
        if option in route:
            route_obj[option] = route[option]
    return route_obj


def _validate_rules(rules):
    """Validate the format of policy-based routing rules.

    :param rules: a list of rules or None.
    :raises: AnsibleFilterError if any rule is invalid.
    """
    for rule in rules or []:
        if not isinstance(rule, str):
            raise errors.AnsibleFilterError(
                "Routing policy rules must be defined in string format "
                "for CentOS")


@jinja2.contextfilter
def net_interface_obj(context, name, inventory_hostname=None):
    """Return a dict representation of a network interface.

    The returned dict is compatible with the interfaces_ether_interfaces
    variable in the MichaelRigaert.interfaces role.
    """
    device = net_interface(context, name, inventory_hostname)
    if not device:
        raise errors.AnsibleFilterError(
            "Network interface for network '%s' on host '%s' not found" %
            (name, inventory_hostname))
    ip = net_ip(context, name, inventory_hostname)
    netmask = net_mask(context, name, inventory_hostname)
    gateway = net_gateway(context, name, inventory_hostname)
    if ip is None:
        ip = ''
        gateway = None
        netmask = None
    vlan = net_vlan(context, name, inventory_hostname)
    mtu = net_mtu(context, name, inventory_hostname)
    routes = net_routes(context, name, inventory_hostname)
    if routes:
        routes = [_route_obj(route) for route in routes]
    rules = net_rules(context, name, inventory_hostname)
    bootproto = net_bootproto(context, name, inventory_hostname)
    defroute = net_defroute(context, name, inventory_hostname)
    ethtool_opts = net_ethtool_opts(context, name, inventory_hostname)
    zone = net_zone(context, name, inventory_hostname)
    vip_address = net_vip_address(context, name, inventory_hostname)
    allowed_addresses = [vip_address] if vip_address else None
    _validate_rules(rules)
    interface = {
        'device': device,
        'address': ip,
        'netmask': netmask,
        'gateway': gateway,
        'vlan': vlan,
        'mtu': mtu,
        'route': routes,
        'rules': rules,
        'bootproto': bootproto or 'static',
        'defroute': defroute,
        'ethtool_opts': ethtool_opts,
        'zone': zone,
        'allowed_addresses': allowed_addresses,
        'onboot': 'yes',
    }
    interface = {k: v for k, v in interface.items() if v is not None}
    return interface


@jinja2.contextfilter
def net_bridge_obj(context, name, inventory_hostname=None):
    """Return a dict representation of a network bridge interface.

    The returned dict is compatible with the interfaces_bridge_interfaces
    variable in the MichaelRigaert.interfaces role.
    """
    device = net_interface(context, name, inventory_hostname)
    if not device:
        raise errors.AnsibleFilterError(
            "Network interface for network '%s' on host '%s' not found" %
            (name, inventory_hostname))
    ip = net_ip(context, name, inventory_hostname)
    netmask = net_mask(context, name, inventory_hostname)
    gateway = net_gateway(context, name, inventory_hostname)
    if ip is None:
        ip = ''
        gateway = None
        netmask = None
    vlan = net_vlan(context, name, inventory_hostname)
    mtu = net_mtu(context, name, inventory_hostname)
    ports = net_bridge_ports(context, name, inventory_hostname)
    routes = net_routes(context, name, inventory_hostname)
    if routes:
        routes = [_route_obj(route) for route in routes]
    rules = net_rules(context, name, inventory_hostname)
    bootproto = net_bootproto(context, name, inventory_hostname)
    defroute = net_defroute(context, name, inventory_hostname)
    ethtool_opts = net_ethtool_opts(context, name, inventory_hostname)
    zone = net_zone(context, name, inventory_hostname)
    vip_address = net_vip_address(context, name, inventory_hostname)
    allowed_addresses = [vip_address] if vip_address else None
    _validate_rules(rules)
    interface = {
        'device': device,
        'address': ip,
        'netmask': netmask,
        'gateway': gateway,
        'vlan': vlan,
        'mtu': mtu,
        'ports': ports,
        'route': routes,
        'rules': rules,
        'bootproto': bootproto or 'static',
        'defroute': defroute,
        'ethtool_opts': ethtool_opts,
        'zone': zone,
        'allowed_addresses': allowed_addresses,
        'onboot': 'yes',
    }
    interface = {k: v for k, v in interface.items() if v is not None}
    return interface


@jinja2.contextfilter
def net_bond_obj(context, name, inventory_hostname=None):
    """Return a dict representation of a network bond interface.

    The returned dict is compatible with the interfaces_bond_interfaces
    variable in the MichaelRigaert.interfaces role.
    """
    device = net_interface(context, name, inventory_hostname)
    if not device:
        raise errors.AnsibleFilterError(
            "Network interface for network '%s' on host '%s' not found" %
            (name, inventory_hostname))
    ip = net_ip(context, name, inventory_hostname)
    netmask = net_mask(context, name, inventory_hostname)
    gateway = net_gateway(context, name, inventory_hostname)
    if ip is None:
        ip = ''
        gateway = None
        netmask = None
    vlan = net_vlan(context, name, inventory_hostname)
    mtu = net_mtu(context, name, inventory_hostname)
    mode = net_bond_mode(context, name, inventory_hostname)
    slaves = net_bond_slaves(context, name, inventory_hostname)
    miimon = net_bond_miimon(context, name, inventory_hostname)
    updelay = net_bond_updelay(context, name, inventory_hostname)
    downdelay = net_bond_downdelay(context, name, inventory_hostname)
    xmit_hash_policy = net_bond_xmit_hash_policy(context, name,
                                                 inventory_hostname)
    lacp_rate = net_bond_lacp_rate(context, name, inventory_hostname)
    routes = net_routes(context, name, inventory_hostname)
    if routes:
        routes = [_route_obj(route) for route in routes]
    rules = net_rules(context, name, inventory_hostname)
    bootproto = net_bootproto(context, name, inventory_hostname)
    defroute = net_defroute(context, name, inventory_hostname)
    ethtool_opts = net_ethtool_opts(context, name, inventory_hostname)
    zone = net_zone(context, name, inventory_hostname)
    vip_address = net_vip_address(context, name, inventory_hostname)
    allowed_addresses = [vip_address] if vip_address else None
    _validate_rules(rules)
    interface = {
        'device': device,
        'address': ip,
        'netmask': netmask,
        'gateway': gateway,
        'vlan': vlan,
        'mtu': mtu,
        'bond_slaves': slaves,
        'bond_mode': mode,
        'bond_miimon': miimon,
        'bond_updelay': updelay,
        'bond_downdelay': downdelay,
        'bond_xmit_hash_policy': xmit_hash_policy,
        'bond_lacp_rate': lacp_rate,
        'route': routes,
        'rules': rules,
        'bootproto': bootproto or 'static',
        'defroute': defroute,
        'ethtool_opts': ethtool_opts,
        'zone': zone,
        'allowed_addresses': allowed_addresses,
        'onboot': 'yes',
    }
    interface = {k: v for k, v in interface.items() if v is not None}
    return interface


def _net_interface_type(context, name, inventory_hostname):
    """Return a string describing the network interface type.

    Possible types include 'ether', 'bridge', 'bond'.
    """
    bridge_ports = net_bridge_ports(context, name, inventory_hostname)
    bond_slaves = net_bond_slaves(context, name, inventory_hostname)
    if bridge_ports is not None and bond_slaves is not None:
        raise errors.AnsibleFilterError(
            "Network %s on host %s has both bridge ports and bond slaves "
            "defined" %
            (name,
             utils.get_hostvar(context, 'inventory_hostname',
                               inventory_hostname)))
    if bridge_ports is None and bond_slaves is None:
        return 'ether'
    if bridge_ports is not None:
        return 'bridge'
    if bond_slaves is not None:
        return 'bond'


@jinja2.contextfilter
def net_is_ether(context, name, inventory_hostname=None):
    return _net_interface_type(context, name, inventory_hostname) == 'ether'


@jinja2.contextfilter
def net_is_bridge(context, name, inventory_hostname=None):
    return _net_interface_type(context, name, inventory_hostname) == 'bridge'


@jinja2.contextfilter
def net_is_bond(context, name, inventory_hostname=None):
    return _net_interface_type(context, name, inventory_hostname) == 'bond'


@jinja2.contextfilter
def net_is_vlan(context, name, inventory_hostname=None):
    return net_vlan(context, name) is not None


@jinja2.contextfilter
def net_is_vlan_interface(context, name, inventory_hostname=None):
    device = get_and_validate_interface(context, name, inventory_hostname)
    # Use a heuristic to match conventional VLAN names, ending with a
    # period and a numerical extension to an interface name
    return re.match(r"^[a-zA-Z0-9_\-]+\.[1-9][\d]{0,3}$", device)


@jinja2.contextfilter
def net_select_ethers(context, names, inventory_hostname=None):
    return [name for name in names
            if net_is_ether(context, name, inventory_hostname)]


@jinja2.contextfilter
def net_select_bridges(context, names, inventory_hostname=None):
    return [name for name in names
            if net_is_bridge(context, name, inventory_hostname)]


@jinja2.contextfilter
def net_select_bonds(context, names, inventory_hostname=None):
    return [name for name in names
            if net_is_bond(context, name, inventory_hostname)]


@jinja2.contextfilter
def net_select_vlans(context, names, inventory_hostname=None):
    return [name for name in names
            if net_is_vlan(context, name, inventory_hostname)]


@jinja2.contextfilter
def net_select_vlan_interfaces(context, names, inventory_hostname=None):
    return [name for name in names
            if net_is_vlan_interface(context, name, inventory_hostname)]


@jinja2.contextfilter
def net_reject_vlans(context, names, inventory_hostname=None):
    return [name for name in names
            if not net_is_vlan(context, name, inventory_hostname)]


@jinja2.contextfilter
def net_configdrive_network_device(context, name, inventory_hostname=None):
    device = net_interface(context, name, inventory_hostname)
    if not device:
        raise errors.AnsibleFilterError(
            "Network interface for network '%s' on host '%s' not found" %
            (name, inventory_hostname))
    ip = net_ip(context, name, inventory_hostname)
    netmask = net_mask(context, name, inventory_hostname)
    gateway = net_gateway(context, name, inventory_hostname)
    bootproto = net_bootproto(context, name, inventory_hostname)
    mtu = net_mtu(context, name, inventory_hostname)
    interface = {
        'device': device,
        'address': ip,
        'netmask': netmask,
        'gateway': gateway,
        'bootproto': bootproto or 'static',
        'mtu': mtu,
    }
    interface = {k: v for k, v in interface.items() if v is not None}
    return interface


@jinja2.contextfilter
def net_libvirt_network(context, name, inventory_hostname=None):
    """Return a dict which describes the Libvirt network for a network.

    The Libvirt network is in a form accepted by the libvirt-host role.
    """
    interface = net_interface(context, name, inventory_hostname)
    name = net_libvirt_network_name(context, name, inventory_hostname)
    return {
        "name": name,
        "mode": "bridge",
        "bridge": interface,
    }


@jinja2.contextfilter
def net_libvirt_vm_network(context, name, inventory_hostname=None):
    """Return a dict which describes the Libvirt VM's network for a network.

    The Libvirt network is in a form accepted by the libvirt_vm_interfaces
    variable of the libvirt-vm role.
    """
    libvirt_name = net_libvirt_network_name(context, name, inventory_hostname)
    return {
        "network": libvirt_name,
        "net_name": name,
    }


@jinja2.contextfilter
def net_ovs_veths(context, names, inventory_hostname=None):
    """Return a list of virtual Ethernet pairs for OVS.

    The format is as expected by the veth_interfaces variable of the Kayobe
    veth role.
    """
    veths = get_ovs_veths(context, names, inventory_hostname)
    return [
        {
            'device': veth['name'],
            'bootproto': 'static',
            'bridge': veth['bridge'],
            'mtu': veth['mtu'],
            'peer_device': veth['peer'],
            'peer_bootproto': 'static',
            'peer_mtu': veth['mtu'],
            'onboot': 'yes',
        }
        for veth in veths
    ]


def get_filters():
    return {
        'net_attr': net_attr,
        'net_vip_address': net_vip_address,
        'net_fqdn': _make_attr_filter('fqdn'),
        'net_ip': net_ip,
        'net_interface': net_interface,
        'net_cidr': net_cidr,
        'net_mask': net_mask,
        'net_prefix': net_prefix,
        'net_gateway': net_gateway,
        'net_allocation_pool_start': net_allocation_pool_start,
        'net_allocation_pool_end': net_allocation_pool_end,
        'net_inspection_allocation_pool_start': (
            net_inspection_allocation_pool_start),
        'net_inspection_allocation_pool_end': (
            net_inspection_allocation_pool_end),
        'net_inspection_gateway': net_inspection_gateway,
        'net_neutron_allocation_pool_start': net_neutron_allocation_pool_start,
        'net_neutron_allocation_pool_end': net_neutron_allocation_pool_end,
        'net_neutron_gateway': net_neutron_gateway,
        'net_vlan': net_vlan,
        'net_mtu': net_mtu,
        'net_routes': net_routes,
        'net_rules': net_rules,
        'net_physical_network': net_physical_network,
        'net_bootproto': net_bootproto,
        'net_defroute': net_defroute,
        'net_ethtool_opts': net_ethtool_opts,
        'net_zone': net_zone,
        'net_interface_obj': net_interface_obj,
        'net_bridge_obj': net_bridge_obj,
        'net_bond_obj': net_bond_obj,
        'net_is_ether': net_is_ether,
        'net_is_bridge': net_is_bridge,
        'net_is_bond': net_is_bond,
        'net_is_vlan': net_is_vlan,
        'net_select_ethers': net_select_ethers,
        'net_select_bridges': net_select_bridges,
        'net_select_bonds': net_select_bonds,
        'net_select_vlans': net_select_vlans,
        'net_reject_vlans': net_reject_vlans,
        'net_configdrive_network_device': net_configdrive_network_device,
        'net_libvirt_network_name': net_libvirt_network_name,
        'net_libvirt_network': net_libvirt_network,
        'net_libvirt_vm_network': net_libvirt_vm_network,
        'net_ovs_veths': net_ovs_veths,
    }



# Copyright (c) 2018 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import os.path

from kayobe.tests.molecule import utils

import pytest
import testinfra.utils.ansible_runner


testinfra_hosts = testinfra.utils.ansible_runner.AnsibleRunner(
    os.environ['MOLECULE_INVENTORY_FILE']).get_hosts('all')


@pytest.mark.parametrize(
    'path',
    ['fluentd/filter',
     'fluentd/input',
     'fluentd/output'])
def test_service_config_directory(host, path):
    path = os.path.join('/etc/kolla/config', path)
    utils.test_directory(host, path)


@pytest.mark.parametrize(
    'path',
    ['aodh',
     'cinder',
     'cloudkitty',
     'designate',
     'glance',
     'grafana',
     'heat',
     'horizon',
     'ironic',
     'kafka',
     'keystone',
     'magnum',
     'manila',
     'mariadb',
     'masakari',
     'monasca',
     'murano',
     'neutron',
     'nova',
     'prometheus',
     'sahara',
     'storm',
     'swift',
     'zookeeper'])
def test_service_config_directory_absent(host, path):
    path = os.path.join('/etc/kolla/config', path)
    utils.test_path_absent(host, path)



#!/usr/bin/python3

# Copyright (c) 2017 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy

from ansible.module_utils.basic import *
from ansible.module_utils.openstack import *

# Store a list of import errors to report to the user.
IMPORT_ERRORS = []
try:
    import ironic_inspector_client
except Exception as e:
    IMPORT_ERRORS.append(e)
try:
    import openstack
except Exception as e:
    IMPORT_ERRORS.append(e)


DOCUMENTATION = """
module: os_ironic_inspector_rule
short_description: Create or destroy an Ironic Inspector rule.
author: "Mark Goddard <mark@stackhpc.com>"
extends_documentation_fragment: openstack
description:
  - Create or destroy an Ironic inspector rule.
options:
  state:
    description:
      - State of the rule
    choices: ["present", "absent"]
  uuid:
    description:
      - Globally unique identifier for the rule.
    required: false
  description:
    description:
      - Description for the rule.
    required: false
  conditions:
    description:
      - List of conditions that must be met in order to apply the rule.
    required: true
  actions:
    description:
      - List of actions to be taken when the conditions are met.
    required: true
"""

EXAMPLES = """
# Ensure that an inspector rule exists.
os_ironic_inspector_rule:
  cloud: "openstack"
  state: present
  uuid: "d44666e1-35b3-4f6b-acb0-88ab7052da69"
  description: Set IPMI username in driver_info if not set
  conditions:
    - field: "node://driver_info.ipmi_username"
      op: "is-empty"
  actions:
    - action: "set-attribute"
      path: "driver_info/ipmi_username"
      value: "root"
"""


def _build_client(module, cloud):
    """Create and return an Ironic inspector client."""
    # Ensure the requested API version is supported.
    # API 1.14 is the latest API version available in Rocky.
    api_version = (1, 14)
    client = ironic_inspector_client.v1.ClientV1(
        inspector_url=module.params['inspector_url'],
        interface=module.params['interface'],
        session=cloud.session, region_name=module.params['region_name'],
        api_version=api_version)
    return client


def _ensure_rule_present(module, client):
    """Ensure that an inspector rule is present."""
    if module.params['uuid']:
        try:
            rule = client.rules.get(module.params['uuid'])
        except ironic_inspector_client.ClientError as e:
            if e.response.status_code != 404:
                module.fail_json(msg="Failed retrieving Inspector rule %s: %s"
                                 % (module.params['uuid'], repr(e)))
        else:
            # Check whether the rule differs from the request.
            keys = ('conditions', 'actions', 'description')
            for key in keys:
                expected = module.params[key]
                if key == 'conditions':
                    # Rules returned from the API include default values in the
                    # conditions that may not be in the requested rule. Apply
                    # defaults to allow the comparison to succeed.
                    expected = copy.deepcopy(expected)
                    for condition in expected:
                        condition.setdefault('invert', False)
                        condition.setdefault('multiple', 'any')
                if rule[key] != expected:
                    break
            else:
                # Nothing to do - rule exists and is as requested.
                return False
            # Rule differs - delete it before recreating.
            _ensure_rule_absent(module, client)

    client.rules.create(module.params['conditions'], module.params['actions'],
                        module.params['uuid'], module.params['description'])
    return True


def _ensure_rule_absent(module, client):
    """Ensure that an inspector rule is absent."""
    if not module.params['uuid']:
        module.fail_json(msg="UUID is required to ensure rules are absent")
    try:
        client.rules.delete(module.params['uuid'])
    except ironic_inspector_client.ClientError as e:
        # If the rule does not exist, no problem and no change.
        if e.response.status_code == 404:
            return False
        module.fail_json(msg="Failed retrieving Inspector rule %s: %s"
                         % (module.params['uuid'], repr(e)))
    return True


def main():
    argument_spec = openstack_full_argument_spec(
        conditions=dict(type='list', required=True),
        actions=dict(type='list', required=True),
        description=dict(required=False),
        uuid=dict(required=False),
        state=dict(required=False, default='present',
                   choices=['present', 'absent']),
        inspector_url=dict(required=False),
    )
    module_kwargs = openstack_module_kwargs()
    module = AnsibleModule(argument_spec, **module_kwargs)

    # Fail if there were any exceptions when importing modules.
    if IMPORT_ERRORS:
        module.fail_json(msg="Import errors: %s" %
                         ", ".join([repr(e) for e in IMPORT_ERRORS]))

    if (module.params['auth_type'] in [None, 'None'] and
            module.params['inspector_url'] is None):
        module.fail_json(msg="Authentication appears disabled, please "
                             "define an inspector_url parameter")

    if (module.params['inspector_url'] and
            module.params['auth_type'] in [None, 'None']):
        module.params['auth'] = dict(
            endpoint=module.params['inspector_url']
        )

    sdk, cloud = openstack_cloud_from_module(module)
    try:
        client = _build_client(module, cloud)
        if module.params["state"] == "present":
            changed = _ensure_rule_present(module, client)
        else:
            changed = _ensure_rule_absent(module, client)
    except Exception as e:
        module.fail_json(msg="Failed to configure Ironic Inspector rule: %s" %
                         repr(e))
    else:
        module.exit_json(changed=changed)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""
Script to build a Swift ring from a declarative YAML configuration. This has
been built via a script to avoid repeated 'docker exec' commands which could
take a long time.

Usage:

python swift-ring-builder.py <config file path> <build path> <service name>

Example:

python swift-ring-builder.py /path/to/config.yml /path/to/builds object

Example configuration format:

---
part_power: 10
replication_count: 3
min_part_hours: 1
hosts:
  - host: swift1
    region: 1
    zone: 1
    ip: 10.0.0.1
    port: 6001
    replication_ip: 10.1.0.1
    replication_port: 6001
    devices:
      - device: /dev/sdb
        weight: 100
      - device: /dev/sdc
        weight: 100
"""
import subprocess
import sys

import yaml


class RingBuilder(object):
    """Helper class for building Swift rings."""

    def __init__(self, build_path, service_name):
        self.build_path = build_path
        self.service_name = service_name

    def get_base_command(self):
        return [
            'swift-ring-builder',
            '%s/%s.builder' % (self.build_path, self.service_name),
        ]

    def create(self, part_power, replication_count, min_part_hours):
        cmd = self.get_base_command()
        cmd += [
            'create',
            "{}".format(part_power),
            "{}".format(replication_count),
            "{}".format(min_part_hours),
        ]
        try:
            subprocess.check_call(cmd)
        except subprocess.CalledProcessError:
            print("Failed to create %s ring" % self.service_name)
            sys.exit(1)

    def add_device(self, host, device):
        cmd = self.get_base_command()
        cmd += [
            'add',
            '--region', "{}".format(host['region']),
            '--zone', "{}".format(host['zone']),
            '--ip', host['ip'],
            '--port', "{}".format(host['port']),
            '--replication-ip', host['replication_ip'],
            '--replication-port', "{}".format(host['replication_port']),
            '--device', device['device'],
            '--weight', "{}".format(device['weight']),
        ]
        try:
            subprocess.check_call(cmd)
        except subprocess.CalledProcessError:
            print("Failed to add device %s on host %s to %s ring" %
                  (host['host'], device['device'], self.service_name))
            sys.exit(1)

    def rebalance(self):
        cmd = self.get_base_command()
        cmd += [
            'rebalance',
        ]
        try:
            subprocess.check_call(cmd)
        except subprocess.CalledProcessError:
            print("Failed to rebalance %s ring" % self.service_name)
            sys.exit(1)


def build_rings(config, build_path, service_name):
    builder = RingBuilder(build_path, service_name)
    builder.create(config['part_power'], config['replication_count'],
                   config['min_part_hours'])
    for host in config['hosts']:
        devices = host['devices']
        # If no devices are present for this host, this will be None.
        if devices is None:
            continue
        for device in devices:
            builder.add_device(host, device)
    builder.rebalance()


def main():
    if len(sys.argv) != 4:
        raise Exception("Usage: {0} <config file path> <build path> "
                        "<service name>")
    config_path = sys.argv[1]
    build_path = sys.argv[2]
    service_name = sys.argv[3]
    with open(config_path) as f:
        config = yaml.safe_load(f)
    build_rings(config, build_path, service_name)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

# Copyright 2015 Sam Yaple
# Copyright 2017 99Cloud Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This file has been adapted from the merge_configs action plugin in Kolla
# Ansible.
# https://opendev.org/openstack/kolla-ansible/src/branch/master/ansible/action_plugins/merge_configs.py

import collections
import os
import shutil
import tempfile

from ansible import constants
from ansible.plugins import action
from io import StringIO

from oslo_config import iniparser

_ORPHAN_SECTION = 'TEMPORARY_ORPHAN_VARIABLE_SECTION'

DOCUMENTATION = '''
---
module: merge_configs
short_description: Merge ini-style configs
description:
     - ConfigParser is used to merge several ini-style configs into one
options:
  dest:
    description:
      - The destination file name
    required: True
    type: str
  sources:
    description:
      - A list of files on the destination node to merge together
    default: None
    required: True
    type: str
  whitespace:
    description:
      - Whether whitespace characters should be used around equal signs
    default: True
    required: False
    type: bool
author: Sam Yaple
'''

EXAMPLES = '''
Merge multiple configs:

- hosts: database
  tasks:
    - name: Merge configs
      merge_configs:
        sources:
          - "/tmp/config_1.cnf"
          - "/tmp/config_2.cnf"
          - "/tmp/config_3.cnf"
        dest:
          - "/etc/mysql/my.cnf"
'''


class OverrideConfigParser(iniparser.BaseParser):

    def __init__(self, whitespace=True):
        self._cur_sections = collections.OrderedDict()
        self._sections = collections.OrderedDict()
        self._cur_section = None
        self._whitespace = ' ' if whitespace else ''

    def assignment(self, key, value):
        if self._cur_section is None:
            self.new_section(_ORPHAN_SECTION)
        cur_value = self._cur_section.get(key)
        if len(value) == 1 and value[0] == '':
            value = []
        if not cur_value:
            self._cur_section[key] = [value]
        else:
            self._cur_section[key].append(value)

    def parse(self, lineiter):
        self._cur_sections = collections.OrderedDict()
        self._cur_section = None
        super(OverrideConfigParser, self).parse(lineiter)

        # merge _cur_sections into _sections
        for section, values in self._cur_sections.items():
            if section not in self._sections:
                self._sections[section] = collections.OrderedDict()
            for key, value in values.items():
                self._sections[section][key] = value

    def new_section(self, section):
        cur_section = self._cur_sections.get(section)
        if not cur_section:
            cur_section = collections.OrderedDict()
            self._cur_sections[section] = cur_section
        self._cur_section = cur_section
        return cur_section

    def write(self, fp):
        def write_key_value(key, values):
            for v in values:
                if not v:
                    fp.write('{key}{ws}=\n'.format(
                        key=key, ws=self._whitespace))
                for index, value in enumerate(v):
                    if index == 0:
                        fp.write('{key}{ws}={ws}{value}\n'.format(
                            key=key,
                            ws=self._whitespace,
                            value=value))
                    else:
                        # We want additional values to be written out under the
                        # first value with the same indentation, like this:
                        # key = value1
                        #       value2
                        indent_size = len(key) + len(self._whitespace) * 2 + 1
                        ws_indent = ' ' * indent_size
                        fp.write('{ws_indent}{value}\n'.format(
                            ws_indent=ws_indent,
                            value=value))

        def write_section(section):
            for key, values in section.items():
                write_key_value(key, values)

        for section in self._sections:
            if section != _ORPHAN_SECTION:
                fp.write('[{}]\n'.format(section))
            write_section(self._sections[section])
            fp.write('\n')


class ActionModule(action.ActionBase):

    TRANSFERS_FILES = True

    def read_config(self, source, config):
        # Only use config if present
        if os.access(source, os.R_OK):
            with open(source, 'r') as f:
                template_data = f.read()

            # set search path to mimic 'template' module behavior
            searchpath = [
                self._loader._basedir,
                os.path.join(self._loader._basedir, 'templates'),
                os.path.dirname(source),
            ]
            self._templar.environment.loader.searchpath = searchpath

            result = self._templar.template(template_data)
            fakefile = StringIO(result)
            config.parse(fakefile)
            fakefile.close()

    def run(self, tmp=None, task_vars=None):

        result = super(ActionModule, self).run(tmp, task_vars)
        del tmp  # not used

        sources = self._task.args.get('sources', None)

        if not isinstance(sources, list):
            sources = [sources]

        config = OverrideConfigParser(
            whitespace=self._task.args.get('whitespace', True))

        for source in sources:
            self.read_config(source, config)

        # Dump configparser to string via an emulated file

        fakefile = StringIO()
        config.write(fakefile)
        full_source = fakefile.getvalue()
        fakefile.close()

        local_tempdir = tempfile.mkdtemp(dir=constants.DEFAULT_LOCAL_TMP)

        try:
            result_file = os.path.join(local_tempdir, 'source')
            with open(result_file, 'w') as f:
                f.write(full_source)

            new_task = self._task.copy()
            new_task.args.pop('sources', None)
            new_task.args.pop('whitespace', None)

            new_task.args.update(
                dict(
                    src=result_file
                )
            )

            copy_action = self._shared_loader_obj.action_loader.get(
                'copy',
                task=new_task,
                connection=self._connection,
                play_context=self._play_context,
                loader=self._loader,
                templar=self._templar,
                shared_loader_obj=self._shared_loader_obj)
            result.update(copy_action.run(task_vars=task_vars))
        finally:
            shutil.rmtree(local_tempdir)
        return result
